""" 
    Harry is a small tool for measuring the similarity of strings.  The
    module supports common distance and kernel functions for strings as 
    well as some exotic similarity measures.

    The Python module of Harry uses largely the same parameters as the
    command-line tool. See manual page harry(1) for more information.
    Copyright (C) 2013-2015 Konrad Rieck (konrad@mlsec.org)
"""

import os
import subprocess as sp
import shlex
import urllib.request, urllib.parse, urllib.error
import struct
import warnings

# Construct path for Harry tool
__tool = os.path.join("/usr/bin", "harry")


def __run_harry(strs, opts):
    """ 
    Internal function to call the tool Harry

    Parameters
    ----------
    strs (list)               List of strings to compare.
    opts (str)                Command-line options for Harry

    Returns
    -------
    out (ndarray)             Matrix of similarity values
    """

    # Check for presence of command-line tool
    if not os.path.exists(__tool):
        raise Exception("Could not find tool at %s" % __tool)

    # Decode special characters and prepare input
    strs = list(map(urllib.parse.quote, strs))
    stdin = '\n'.join(strs)

    # Input: "-"  Read strings from standard input
    # Output: "="  Write raw matrix to standard output
    cmd = "%s %s - =" % (__tool, opts)
    args = shlex.split(cmd)
    p = sp.Popen(args, stdout=sp.PIPE, stdin=sp.PIPE, stderr=None)
    stdout, stderr = p.communicate(input=stdin.encode())
    if p.returncode != 0:
        raise Exception("Error while executing %s. See above." % __tool)

    # No output available
    if len(stdout) == 0:
        return None

    # Unpack dimensions of matrix and size of float
    rows, cols, fsize = struct.unpack("III", stdout[0:12])

    try:
        # Check for Numpy
        import numpy

        # Generate matrix form buffer and reshape it
        if fsize == 4:
            mat = numpy.frombuffer(stdout[12:], dtype=numpy.float32)
        elif fsize == 8:
            mat = numpy.frombuffer(stdout[12:], dtype=numpy.float64)
        else:
            raise Exception("Unsupported float type")

        mat = numpy.reshape(mat, (rows, cols))

    except ImportError:
        # Numpy is not available
        warnings.warn("The package numpy is recommended for using Harry")

        if fsize == 4:
            fmt = 'f'
        elif fsize == 8:
            fmt = 'd'
        else:
            raise Exception("Unsupported float type");

        # Generate matrix using list of lists :(
        mat = []
        k = 12
        for i in range(rows):
            row = []
            for j in range(cols):
                row.append(struct.unpack(fmt, stdout[k:k + fsize])[0])
                k += fsize
            mat.append(row)

    return mat


def compare(x, y=None, **kwargs):
    """ 
    Compare strings using a similarity measure

    Parameters
    ----------
    x (list)                  List of strings to compare   
    y (list)                  List of strings to compare [optional]
    
    Parameters (Measure options)
    ----------
    measure (str)             Set similarity measure.
    granularity (str)         Set granularity: bytes, bits, tokens.
    token_delim (str)         Set delimiters for tokens.
    num_threads (int)         Set number of threads.
    cache_size (int)          Set size of cache in megabytes.
    global_cache (bool)       Enable global cache.
    col_range (str)           Set the column range (x) of strings.
    row_range (str)           Set the row range (y) of strings.
    split (str)               Split matrix into blocks and compute one.

    Parameters (Generic options)
    ----------
    config_file (str)         Set configuration file.
    verbose (bool)            Increase verbosity.
    log_line (bool)           Print a log line every minutes.
    quiet (bool)              Be quiet during processing.
    print_measures (bool)     Print list of similarity measures.
    print_config (bool)       Print the current configuration.
    print_defaults (bool)     Print the default configuration.
    version (bool)            Print version and copyright.
    help (bool)               Print help screen.

    Returns
    -------
    out (ndarray)             Matrix of similarity values
    """

    keyargs = set([
        "measure", "granularity", "token_delim", "num_threads", "cache_size", 
        "global_cache", "col_range", "row_range", "split", "config_file", 
        "verbose", "log_line", "quiet", "print_measures", "print_config", 
        "print_defaults", "version", "help"
    ])

    # Process options
    opts = "--decode_str"
    for (k, v) in list(kwargs.items()):
        if k not in keyargs:
            raise Exception("Unknown argument '%s'" % k)
        if type(v) is bool:
            opts += " --%s" % k
        elif type(v) is float or type(v) is int:
            opts += " --%s=%g" % (k, v)
        else:
            opts += " --%s=%s" % (k, v)

    # Fix incorrect usage
    x = x if type(x) is list else [x]

    if y:
        # Fix incorrect usage again
        y = y if type(y) is list else [y]

        # We merge the lists and use index ranges
        opts += " --row_range :%d --col_range %d:" % (len(x), len(x))
        x = x + y

    if len(x) < 50:
        warnings.warn("Harry is not efficient on small sets of strings.")

    return __run_harry(x, opts)


def print_measures():
    """
    Print list of similarity measures supported by Harry
    """
    __run_harry([], "--print_measures")


def print_defaults():
    """
    Print default configuration of Harry
    """
    __run_harry([], "--print_defaults")
